#include "precomp.h"

CCP210x::CCP210x(ICP210xDevices* pDevices, DWORD dwDeviceNum, HRESULT* phr)
    : m_pDevices(pDevices), m_pTypeInfo(nullptr), m_hDevice(INVALID_HANDLE_VALUE), m_lRefCount(1)
{
    m_pDevices->AddRef();

    ITypeLib* pTlb = nullptr;
    if(SUCCEEDED((*phr = LoadRegTypeLib(LIBID_CP210xExLib, 1, 0, LOCALE_USER_DEFAULT, &pTlb))))
    {
        if(SUCCEEDED((*phr = pTlb->GetTypeInfoOfGuid(IID_ICP210x, &m_pTypeInfo))))
        {
            *phr = ::StatusToHresult(::CP210x_Open(dwDeviceNum, &m_hDevice));
        }
        pTlb->Release();
    }
}

CCP210x::~CCP210x()
{
    if(INVALID_HANDLE_VALUE != m_hDevice)
        ::CP210x_Close(m_hDevice);

    if(m_pTypeInfo)
        m_pTypeInfo->Release();

    m_pDevices->Release();
}

IFACEMETHODIMP
CCP210x::QueryInterface(REFIID riid, void** ppvObject)
{
    __try
    {
        *ppvObject = nullptr;

        if(IsEqualIID(riid, IID_IUnknown))
            *ppvObject = GetUnknown();
        else if(IsEqualIID(riid, IID_IDispatch))
            *ppvObject = static_cast<IDispatch*>(this);
        else if(IsEqualIID(riid, IID_ICP210x))
            *ppvObject = static_cast<ICP210x*>(this);
        else if(IsEqualIID(riid, IID_IProvideClassInfo))
            *ppvObject = static_cast<IProvideClassInfo*>(this);
        else if(IsEqualIID(riid, IID_ISupportErrorInfo))
            *ppvObject = static_cast<ISupportErrorInfo*>(this);

        if(*ppvObject)
        {
            static_cast<IUnknown*>(*ppvObject)->AddRef();
            return(S_OK);
        }
    }
    __except((GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION) ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
    {
        return(E_INVALIDARG);
    }

    return(E_NOINTERFACE);
}

IFACEMETHODIMP_(ULONG)
CCP210x::AddRef()
{
    return(InterlockedIncrement(&m_lRefCount));
}

IFACEMETHODIMP_(ULONG)
CCP210x::Release()
{
    ULONG ulResult;

    if(0 == (ulResult = InterlockedDecrement(&m_lRefCount)))
        delete this;

    return(ulResult);
}

IFACEMETHODIMP
CCP210x::GetTypeInfoCount(UINT* pctinfo)
{
    if(IsBadWritePtrT(pctinfo))
        return(E_POINTER);

    *pctinfo = 1;
    return(S_OK);
}

IFACEMETHODIMP
CCP210x::GetTypeInfo(UINT iTInfo, LCID, ITypeInfo** ppTInfo)
{
    if(IsBadWritePtrT(ppTInfo))
        return(E_POINTER);

    if(iTInfo)
    {
        *ppTInfo = nullptr;
        return(DISP_E_BADINDEX);
    }

    *ppTInfo = m_pTypeInfo;
    (*ppTInfo)->AddRef();
    return(S_OK);
}

IFACEMETHODIMP
CCP210x::GetIDsOfNames(REFIID, LPOLESTR* rgszNames, UINT cNames, LCID, DISPID* rgDispId)
{
    return(m_pTypeInfo->GetIDsOfNames(rgszNames, cNames, rgDispId));
}

IFACEMETHODIMP
CCP210x::Invoke(DISPID dispIdMember, REFIID, LCID lcid, WORD wFlags, DISPPARAMS* pDispParams, VARIANT* pVarResult, EXCEPINFO* pExcepInfo, UINT* puArgErr)
{
    CThreadLocale loc(lcid);
    UINT uArgErr;
    return(m_pTypeInfo->Invoke(static_cast<ICP210x*>(this), dispIdMember, wFlags, pDispParams, pVarResult, pExcepInfo, (puArgErr) ? puArgErr : &uArgErr));
}

IFACEMETHODIMP
CCP210x::get_Devices(IDispatch** ppDevices)
{
    const HRESULT hr = m_pDevices->QueryInterface(IID_IDispatch, reinterpret_cast<void**>(ppDevices));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_DEVICES, hr));
}

IFACEMETHODIMP
CCP210x::GetPartNumber(enum PartNumbers* pPartNum)
{
    CP210x_STATUS   stat;
    BYTE            bNum;
    
    if(IsBadWritePtrT(pPartNum))
        return(RaiseCOMException(DISPID_CP210X_GETPARTNUMBER, E_POINTER));

    *pPartNum = static_cast<enum PartNumbers>(0);

    if(CP210x_SUCCESS != (stat = ::CP210x_GetPartNumber(m_hDevice, &bNum)))
        return(RaiseCOMException(DISPID_CP210X_GETPARTNUMBER, ::StatusToHresult(stat)));

    *pPartNum = static_cast<enum PartNumbers>(bNum);
    return(S_OK);    
}

IFACEMETHODIMP
CCP210x::SetVid(long VID)
{
    const HRESULT hr = ::StatusToHresult(::CP210x_SetVid(m_hDevice, static_cast<WORD>(VID)));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETVID, hr));
}

IFACEMETHODIMP
CCP210x::SetPid(long PID)
{
    const HRESULT hr = ::StatusToHresult(::CP210x_SetPid(m_hDevice, static_cast<WORD>(PID)));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETPID, hr));
}

IFACEMETHODIMP
CCP210x::SetProductString(BSTR strProduct)
{
    const HRESULT hr = ::StatusToHresult(::CP210x_SetProductString(m_hDevice, strProduct, static_cast<BYTE>(SysStringByteLen(strProduct)), FALSE));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETPRODUCTSTRING, hr));
}

IFACEMETHODIMP
CCP210x::SetSerialNumber(BSTR strSerialNumber)
{
    const HRESULT hr = ::StatusToHresult(::CP210x_SetSerialNumber(m_hDevice, strSerialNumber, static_cast<BYTE>(SysStringLen(strSerialNumber)), FALSE));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETSERIALNUMBER, hr));
}

IFACEMETHODIMP
CCP210x::SetSelfPower(VARIANT_BOOL SelfPower)
{
    const HRESULT hr = ::StatusToHresult(::CP210x_SetSelfPower(m_hDevice, (SelfPower != VARIANT_FALSE)));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETSELFPOWER, hr));
}

IFACEMETHODIMP
CCP210x::SetMaxPower(long MaxPower)
{
    if(static_cast<UINT_PTR>(MaxPower) > CP210x_MAX_MAXPOWER)
        return(RaiseCOMException(DISPID_CP210X_SETMAXPOWER, E_INVALIDARG));

    const  HRESULT hr = ::StatusToHresult(::CP210x_SetMaxPower(m_hDevice, static_cast<BYTE>(MaxPower)));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETMAXPOWER, hr));
}

IFACEMETHODIMP
CCP210x::SetDeviceVersion(long lVersion)
{
    const HRESULT hr = ::StatusToHresult(::CP210x_SetDeviceVersion(m_hDevice, static_cast<WORD>(lVersion)));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETDEVICEVERSION, hr));
}

IFACEMETHODIMP
CCP210x::SetLockValue()
{
    const HRESULT hr = ::StatusToHresult(::CP210x_SetLockValue(m_hDevice));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETLOCKVALUE, hr));
}

IFACEMETHODIMP
CCP210x::GetDeviceProductString(BSTR* ProductString)
{
    CP210x_STATUS   stat;
    WCHAR           wszTemp[CP210x_MAX_PRODUCT_STRLEN + 1];
    BYTE            bLength = _countof(wszTemp);
    
    if(IsBadWritePtrT(ProductString))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEPRODUCTSTRING, E_POINTER));

    *ProductString = nullptr;

    if(CP210x_SUCCESS != (stat = ::CP210x_GetDeviceProductString(m_hDevice, wszTemp, &bLength, FALSE)))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEPRODUCTSTRING, ::StatusToHresult(stat)));

    if(nullptr == (*ProductString = SysAllocStringLen(wszTemp, bLength)))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEPRODUCTSTRING, E_OUTOFMEMORY));

    return(S_OK);
}

IFACEMETHODIMP
CCP210x::GetDeviceSerialNumber(BSTR* SerNum)
{
    CP210x_STATUS   stat;
    WCHAR           wszTemp[CP210x_MAX_SERIAL_STRLEN + 1];
    BYTE            bLength = _countof(wszTemp);

    if(IsBadWritePtrT(SerNum))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICESERIALNUMBER, E_POINTER));

    *SerNum = nullptr;

    if(CP210x_SUCCESS != (stat = ::CP210x_GetDeviceSerialNumber(m_hDevice, wszTemp, &bLength, FALSE)))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICESERIALNUMBER, ::StatusToHresult(stat)));

    if(nullptr == (*SerNum = SysAllocStringLen(wszTemp, bLength)))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICESERIALNUMBER, E_OUTOFMEMORY));

    return(S_OK);
}

IFACEMETHODIMP
CCP210x::GetDeviceVid(long* VID)
{
    CP210x_STATUS   stat;
    WORD            w;

    if(IsBadWritePtrT(VID))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEVID, E_POINTER));

    *VID = 0x0000;

    if(CP210x_SUCCESS != (stat = ::CP210x_GetDeviceVid(m_hDevice, &w)))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEVID, ::StatusToHresult(stat)));

    *VID = w;
    return(S_OK);
}

IFACEMETHODIMP
CCP210x::GetDevicePid(long* PID)
{
    CP210x_STATUS   stat;
    WORD            w;

    if(IsBadWritePtrT(PID))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEPID, E_POINTER));

    *PID = 0x0000;

    if(CP210x_SUCCESS != (stat = ::CP210x_GetDevicePid(m_hDevice, &w)))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEPID, ::StatusToHresult(stat)));

    *PID = w;   
    return(S_OK);
}

IFACEMETHODIMP
CCP210x::GetSelfPower(VARIANT_BOOL* SelfPower)
{
    CP210x_STATUS   stat;
    BOOL            b;

    if(IsBadWritePtrT(SelfPower))
        return(RaiseCOMException(DISPID_CP210X_GETSELFPOWER, E_POINTER));

    *SelfPower = VARIANT_FALSE;

    if(CP210x_SUCCESS != (stat = ::CP210x_GetSelfPower(m_hDevice, &b)))
        return(RaiseCOMException(DISPID_CP210X_GETSELFPOWER, ::StatusToHresult(stat)));

    *SelfPower = (b) ? VARIANT_TRUE : VARIANT_FALSE;
    return(S_OK);
}

IFACEMETHODIMP
CCP210x::GetMaxPower(long* MaxPower)
{
    CP210x_STATUS   stat;
    BYTE            b;

    if(IsBadWritePtrT(MaxPower))
        return(RaiseCOMException(DISPID_CP210X_GETMAXPOWER, E_POINTER));

    *MaxPower = CP210x_MAX_MAXPOWER;

    if(CP210x_SUCCESS != (stat = ::CP210x_GetMaxPower(m_hDevice, &b)))
        return(RaiseCOMException(DISPID_CP210X_GETMAXPOWER, ::StatusToHresult(stat)));

    *MaxPower = b;
    return(S_OK);
}

IFACEMETHODIMP
CCP210x::GetDeviceVersion(long* wVersion)
{
    CP210x_STATUS   stat;
    WORD            w;

    if(IsBadWritePtrT(wVersion))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEVERSION, E_POINTER));

    *wVersion = 0;

    if(CP210x_SUCCESS != (stat = ::CP210x_GetDeviceVersion(m_hDevice, &w)))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEVERSION, ::StatusToHresult(stat)));

    *wVersion = w;
    return(S_OK);
}

IFACEMETHODIMP
CCP210x::GetLockValue(VARIANT_BOOL* LockVal)
{
    CP210x_STATUS   stat;
    BYTE            b;

    if(IsBadWritePtrT(LockVal))
        return(RaiseCOMException(DISPID_CP210X_GETLOCKVALUE, E_POINTER));

    *LockVal = VARIANT_TRUE;

    if(CP210x_SUCCESS != (stat = ::CP210x_GetLockValue(m_hDevice, &b)))
        return(RaiseCOMException(DISPID_CP210X_GETLOCKVALUE, ::StatusToHresult(stat)));

    *LockVal = (b) ? VARIANT_TRUE : VARIANT_FALSE;
    return(S_OK);
}

IFACEMETHODIMP
CCP210x::Reset()
{
    const CP210x_STATUS stat = ::CP210x_Reset(m_hDevice);
    // CP210x_Reset returns CP210x_DEVICE_IO_FAILED even on success!
    const HRESULT hr = ::StatusToHresult((stat == CP210x_DEVICE_IO_FAILED) ? CP210x_SUCCESS : stat);
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_RESET, hr));
}

IFACEMETHODIMP
CCP210x::CreateHexFile(BSTR bstrFileName)
{
    PSTR    pstr;
    int     nLen;

    __try
    {
        nLen = WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS, bstrFileName, -1, nullptr, 0, nullptr, nullptr);
    }
    __except((GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION) ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
    {
        return(RaiseCOMException(DISPID_CP210X_CREATEHEXFILE, E_INVALIDARG));
    }

    if(!nLen++)
        return(RaiseCOMException(DISPID_CP210X_CREATEHEXFILE, E_INVALIDARG));

    __try
    {
        pstr = static_cast<PSTR>(_alloca(nLen * sizeof(*pstr)));
    }
    __except((GetExceptionCode() == STATUS_STACK_OVERFLOW) ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
    {
        _resetstkoflw();
        return(RaiseCOMException(DISPID_CP210X_CREATEHEXFILE, E_OUTOFMEMORY));
    }

    WideCharToMultiByte(CP_ACP, WC_NO_BEST_FIT_CHARS, bstrFileName, -1, pstr, nLen, nullptr, nullptr);
    const HRESULT hr = ::StatusToHresult(::CP210x_CreateHexFile(m_hDevice, pstr));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_CREATEHEXFILE, hr));    
}

IFACEMETHODIMP
CCP210x::GetBaudRateConfig(SAFEARRAY** ppBaudConfig)
{
    SAFEARRAYBOUND      sab;
    unsigned int        uIndex;
    BAUD_CONFIG*        conf;
    IRecordInfo*        pRecInfo    = nullptr;
    ITypeInfo*          pTypeInfo   = nullptr;
    ITypeLib*           pTlb        = nullptr;
    HRESULT             hr;

    if(IsBadWritePtrT(ppBaudConfig))
        return(RaiseCOMException(DISPID_CP210X_GETBAUDRATECONFIG, E_POINTER));

    *ppBaudConfig = nullptr;

    if(FAILED((hr = m_pTypeInfo->GetContainingTypeLib(&pTlb, &uIndex))))
        return(RaiseCOMException(DISPID_CP210X_GETBAUDRATECONFIG, hr));

    hr = pTlb->GetTypeInfoOfGuid(GUID_BAUDCONFIG, &pTypeInfo);
    pTlb->Release();

    if(FAILED(hr))
        return(RaiseCOMException(DISPID_CP210X_GETBAUDRATECONFIG, hr));

    hr = GetRecordInfoFromTypeInfo(pTypeInfo, &pRecInfo);
    pTypeInfo->Release();

    if(FAILED(hr))
        return(RaiseCOMException(DISPID_CP210X_GETBAUDRATECONFIG, hr));

    sab.cElements   = NUM_BAUD_CONFIGS;
    sab.lLbound     = 0;
    *ppBaudConfig   = SafeArrayCreateEx(VT_RECORD, 1, &sab, pRecInfo);
    pRecInfo->Release();

    if(nullptr == *ppBaudConfig)
        return(RaiseCOMException(DISPID_CP210X_GETBAUDRATECONFIG, E_OUTOFMEMORY));

    if(SUCCEEDED((hr = SafeArrayAccessData(*ppBaudConfig, (void**)&conf))))
    {
        hr = ::StatusToHresult(::CP210x_GetBaudRateConfig(m_hDevice, conf));
        SafeArrayUnaccessData(*ppBaudConfig);
    }

    if(FAILED(hr))
    {
        SafeArrayDestroy(*ppBaudConfig);
        *ppBaudConfig = nullptr;
        RaiseCOMException(DISPID_CP210X_GETBAUDRATECONFIG, hr);
    }

    return(hr);
}

IFACEMETHODIMP
CCP210x::SetBaudRateConfig(SAFEARRAY** ppBaudConfig)
{
    IRecordInfo*    pRecInfo = nullptr;
    BAUD_CONFIG*    conf;
    HRESULT         hr;
    GUID            guid;
    long            lLBound, lUBound, cElements;

    if(IsBadReadPtrT(ppBaudConfig))
        return(RaiseCOMException(DISPID_CP210X_SETBAUDRATECONFIG, E_POINTER));

    if(1 != SafeArrayGetDim(*ppBaudConfig))
        return(RaiseCOMException(DISPID_CP210X_SETBAUDRATECONFIG, E_INVALIDARG));

    if(FAILED((hr = SafeArrayGetLBound(*ppBaudConfig, 1, &lLBound))))
        return(RaiseCOMException(DISPID_CP210X_SETBAUDRATECONFIG, hr));
        
    if(FAILED((hr = SafeArrayGetUBound(*ppBaudConfig, 1, &lUBound))))
        return(RaiseCOMException(DISPID_CP210X_SETBAUDRATECONFIG, hr));
    
    cElements = lUBound - lLBound + 1;
    
    if(cElements != NUM_BAUD_CONFIGS)
        return(RaiseCOMException(DISPID_CP210X_SETBAUDRATECONFIG, E_INVALIDARG));
    
    if(SUCCEEDED((hr = SafeArrayGetRecordInfo(*ppBaudConfig, &pRecInfo))))
    {
        if(SUCCEEDED((hr = pRecInfo->GetGuid(&guid))))
        {
            if(!IsEqualGUID(guid, GUID_BAUDCONFIG))
            {
                hr = E_INVALIDARG;
            }
            else if(SUCCEEDED((hr = SafeArrayAccessData(*ppBaudConfig, (void**)&conf))))
            {
                hr = ::StatusToHresult(::CP210x_SetBaudRateConfig(m_hDevice, conf));
                SafeArrayUnaccessData(*ppBaudConfig);
            }
        }        
        pRecInfo->Release();
    }
    
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETBAUDRATECONFIG, hr));    
}

IFACEMETHODIMP
CCP210x::GetPortConfig(PortConfig* pConfig)
{
    if(IsBadWritePtrT(pConfig))
        return(RaiseCOMException(DISPID_CP210X_GETPORTCONFIG, E_INVALIDARG));
        
    const HRESULT hr = ::StatusToHresult(::CP210x_GetPortConfig(m_hDevice, reinterpret_cast<PORT_CONFIG*>(pConfig)));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_GETPORTCONFIG, hr));
}

IFACEMETHODIMP
CCP210x::SetPortConfig(PortConfig* pConfig)
{
    if(IsBadReadPtrT(pConfig))
        return(RaiseCOMException(DISPID_CP210X_SETPORTCONFIG, E_INVALIDARG));

    const HRESULT hr = ::StatusToHresult(::CP210x_SetPortConfig(m_hDevice, reinterpret_cast<PORT_CONFIG*>(pConfig)));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETPORTCONFIG, hr));
}

IFACEMETHODIMP
CCP210x::GetFlushBufferConfig(long* pwFlushBufferConfig)
{
    CP210x_STATUS   stat;
    WORD            w = 0;

    if(IsBadWritePtrT(pwFlushBufferConfig))
        return(RaiseCOMException(DISPID_CP210X_GETFLUSHBUFFERCONFIG, E_POINTER));

    *pwFlushBufferConfig = 0;

    if(CP210x_SUCCESS != (stat = ::CP210x_GetFlushBufferConfig(m_hDevice, &w)))
        return(RaiseCOMException(DISPID_CP210X_GETFLUSHBUFFERCONFIG, ::StatusToHresult(stat)));

    *pwFlushBufferConfig = w;
    return(S_OK);
}

IFACEMETHODIMP
CCP210x::SetFlushBufferConfig(long wFlushBufferConfig)
{
    if(static_cast<unsigned long>(wFlushBufferConfig) > MAXWORD)
        return(RaiseCOMException(DISPID_CP210X_SETFLUSHBUFFERCONFIG, E_INVALIDARG));
    
    const HRESULT hr = ::StatusToHresult(::CP210x_SetFlushBufferConfig(m_hDevice, static_cast<WORD>(wFlushBufferConfig)));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETFLUSHBUFFERCONFIG, hr));
}

IFACEMETHODIMP
CCP210x::SetInterfaceString(long InterfaceNumber, BSTR strInterface)
{
    if((static_cast<unsigned long>(InterfaceNumber) > MAXBYTE) || IsBadReadPtrT(strInterface))
        return(RaiseCOMException(DISPID_CP210X_SETINTERFACESTRING, E_INVALIDARG));

    const UINT len = SysStringLen(strInterface);

    if(len > MAXBYTE)
        return(RaiseCOMException(DISPID_CP210X_SETINTERFACESTRING, E_INVALIDARG));

    const HRESULT hr = ::StatusToHresult(::CP210x_SetInterfaceString(m_hDevice, static_cast<BYTE>(InterfaceNumber), strInterface, static_cast<BYTE>(len), FALSE));
    return((SUCCEEDED(hr)) ?  hr : RaiseCOMException(DISPID_CP210X_SETINTERFACESTRING, hr));
}

IFACEMETHODIMP
CCP210x::GetDeviceInterfaceString(long InterfaceNumber, BSTR* strInterface)
{
    HRESULT hr;
    WCHAR   wszTemp[CP2108_MAX_INTERFACE_STRLEN + 1];
    BYTE    len;

    if(IsBadWritePtrT(strInterface))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEINTERFACESTRING, E_POINTER));

    *strInterface = nullptr;

    if(static_cast<unsigned long>(InterfaceNumber) > MAXBYTE)
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEINTERFACESTRING, E_INVALIDARG));

    if(FAILED((hr = ::StatusToHresult(::CP210x_GetDeviceInterfaceString(m_hDevice, static_cast<BYTE>(InterfaceNumber), wszTemp, &len, FALSE)))))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEINTERFACESTRING, hr));

    wszTemp[len] = L'\0';

    if(nullptr == (*strInterface = SysAllocString(wszTemp)))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEINTERFACESTRING, E_OUTOFMEMORY));

    return(S_OK);
}

IFACEMETHODIMP
CCP210x::SetManufacturerString(BSTR strManufacturer)
{
    if(IsBadReadPtrT(strManufacturer))
        return(RaiseCOMException(DISPID_CP210X_SETMANUFACTURERSTRING, E_INVALIDARG));

    const UINT len = SysStringLen(strManufacturer);

    if(len > MAXBYTE)
        return(RaiseCOMException(DISPID_CP210X_SETMANUFACTURERSTRING, E_INVALIDARG));

    const HRESULT hr = ::StatusToHresult(::CP210x_SetManufacturerString(m_hDevice, strManufacturer, static_cast<BYTE>(len), FALSE));
    return((SUCCEEDED(hr)) ?  hr : RaiseCOMException(DISPID_CP210X_SETMANUFACTURERSTRING, hr));
}

IFACEMETHODIMP
CCP210x::GetDeviceManufacturerString(BSTR* strManufacturer)
{
    HRESULT hr;
    WCHAR   wszTemp[CP210x_MAX_MANUFACTURER_STRLEN + 1];
    BYTE    len;

    if(IsBadWritePtrT(strManufacturer))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEMANUFACTURERSTRING, E_POINTER));

    *strManufacturer = nullptr;

    if(FAILED((hr = ::StatusToHresult(::CP210x_GetDeviceManufacturerString(m_hDevice, wszTemp, &len, FALSE)))))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEINTERFACESTRING, hr));

    wszTemp[len] = L'\0';

    if(nullptr == (*strManufacturer = SysAllocString(wszTemp)))
        return(RaiseCOMException(DISPID_CP210X_GETDEVICEMANUFACTURERSTRING, E_OUTOFMEMORY));

    return(S_OK);
}

IFACEMETHODIMP
CCP210x::SetDeviceMode(enum DeviceMode ECI, enum DeviceMode SCI)
{
    if((static_cast<unsigned>(ECI) > modeGpio) || (static_cast<unsigned>(SCI) > modeGpio))
        return(RaiseCOMException(DISPID_CP210X_SETDEVICEMODE, E_INVALIDARG));

    const HRESULT hr = ::StatusToHresult(::CP210x_SetDeviceMode(m_hDevice, static_cast<BYTE>(ECI), static_cast<BYTE>(SCI)));
    return((SUCCEEDED(hr)) ? hr : RaiseCOMException(DISPID_CP210X_SETDEVICEMODE, hr));
}

IFACEMETHODIMP
CCP210x::GetDeviceMode(enum DeviceMode* ECI, enum DeviceMode* SCI)
{
    HRESULT hr = S_OK;
    BYTE bEci = 0, bSci = 0;

    if(!IsBadWritePtrT(ECI))
        *ECI = modeGpio;
    else
        hr = E_POINTER;

    if(!IsBadWritePtrT(SCI))
        *SCI = modeGpio;
    else
        hr = E_POINTER;

    if(SUCCEEDED(hr))
    {
        if(SUCCEEDED(hr = ::StatusToHresult(::CP210x_GetDeviceMode(m_hDevice, &bEci, &bSci))))
        {
            *ECI = static_cast<enum DeviceMode>(bEci);
            *SCI = static_cast<enum DeviceMode>(bSci);
        }
    }

    return(RaiseCOMException(DISPID_CP210X_GETDEVICEMODE, hr));
}

IFACEMETHODIMP
CCP210x::GetClassInfo(ITypeInfo** ppTI)
{
    ITypeLib*   pTlb = nullptr;
    HRESULT     hr;
    UINT        uIndex;

    if(IsBadWritePtrT(ppTI))
        return(E_POINTER);
    
    *ppTI = nullptr;
    
    if(SUCCEEDED(hr = m_pTypeInfo->GetContainingTypeLib(&pTlb, &uIndex)))
    {
        hr = pTlb->GetTypeInfoOfGuid(CLSID_CP210x, ppTI);
        pTlb->Release();
    }
    
    return(hr);
}

IFACEMETHODIMP
CCP210x::InterfaceSupportsErrorInfo(REFIID riid)
{
    __try
    {
        if(IsEqualIID(riid, IID_ICP210x))
            return(S_OK);
    }
    __except((GetExceptionCode() == EXCEPTION_ACCESS_VIOLATION) ? EXCEPTION_EXECUTE_HANDLER : EXCEPTION_CONTINUE_SEARCH)
    {
        return(E_INVALIDARG);
    }
    
    return(S_FALSE);
}
